import ctypes
from collections.abc import Callable
from typing import Self

from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3capi_driver import LinkPtr
from pyautd3.native_methods.utils import ConstantADT, _to_null_terminated_utf8, _validate_ptr
from pyautd3.utils import Duration

from pyautd3_link_ethercrab.native_methods.autd3_capi_link_ethercrab import EtherCrabOption as EtherCrabOption_
from pyautd3_link_ethercrab.native_methods.autd3_capi_link_ethercrab import NativeMethods as LinkEtherCrab
from pyautd3_link_ethercrab.native_methods.autd3_capi_link_ethercrab import Status as Status_

ErrHandlerFunc = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_uint32, ctypes.c_uint8)  # type: ignore[arg-type]


class Status(metaclass=ConstantADT):
    _inner: Status_
    _msg: str

    @classmethod
    def __private_new__(cls: type["Status"], inner: Status_, msg: str) -> "Status":
        ins = super().__new__(cls)
        ins._inner = inner
        ins._msg = msg
        return ins

    def __new__(cls: type["Status"]) -> "Status":
        raise NotImplementedError

    def __repr__(self: Self) -> str:
        return f"{self._msg}"

    def __eq__(self: Self, other: object) -> bool:
        if not isinstance(other, Status):
            return False
        return self._inner == other._inner

    def __hash__(self: Self) -> int:
        return self._inner.__hash__()  # pragma: no cover

    @staticmethod
    def Lost() -> "Status":  # noqa: N802
        return Status.__private_new__(Status_.Lost, "")

    @staticmethod
    def StateChanged() -> "Status":  # noqa: N802
        return Status.__private_new__(Status_.StateChanged, "")

    @staticmethod
    def Error() -> "Status":  # noqa: N802
        return Status.__private_new__(Status_.Error, "")

    @staticmethod
    def Resumed() -> "Status":  # noqa: N802
        return Status.__private_new__(Status_.Resumed, "")


class EtherCrabOption:
    ifname: str | None
    state_check_period: Duration
    sync0_period: Duration
    sync_tolerance: Duration
    sync_timeout: Duration

    def __init__(
        self: Self,
        *,
        ifname: str | None = None,
        state_check_period: Duration | None = None,
        sync0_period: Duration | None = None,
        sync_tolerance: Duration | None = None,
        sync_timeout: Duration | None = None,
    ) -> None:
        self.ifname = ifname
        self.state_check_period = state_check_period or Duration.from_millis(100)
        self.sync0_period = sync0_period or Duration.from_millis(2)
        self.sync_tolerance = sync_tolerance or Duration.from_micros(1)
        self.sync_timeout = sync_timeout or Duration.from_secs(10)

    def _inner(self: Self) -> EtherCrabOption_:
        return EtherCrabOption_(
            _to_null_terminated_utf8(self.ifname) if self.ifname else None,
            self.sync0_period._inner,
            self.state_check_period._inner,
            self.sync_tolerance._inner,
            self.sync_timeout._inner,
        )


class EtherCrab(Link):
    _err_handler: Callable[[int, Status], None]
    _option: EtherCrabOption

    def __init__(self: Self, err_handler: Callable[[int, Status], None], option: EtherCrabOption) -> None:
        super().__init__()
        self._err_handler = err_handler
        self._option = option

    def _resolve(self: Self) -> LinkPtr:
        def callback_native(_context: ctypes.c_void_p, slave: ctypes.c_uint32, status: ctypes.c_uint8) -> None:  # pragma: no cover
            err = bytes(bytearray(128))  # pragma: no cover
            status_ = Status_(int(status))  # pragma: no cover
            LinkEtherCrab().link_ether_crab_status_get_msg(status_, err)  # pragma: no cover
            self._err_handler(int(slave), Status.__private_new__(status_, err.decode("utf-8").rstrip(" \t\r\n\0")))  # pragma: no cover

        self._err_handler_f = ErrHandlerFunc(callback_native)  # pragma: no cover

        return _validate_ptr(  # pragma: no cover
            LinkEtherCrab().link_ether_crab(
                self._err_handler_f,  # type: ignore[arg-type]
                ctypes.c_void_p(0),
                self._option._inner(),
            ),
        )



import pytest

from pyautd3_link_ethercrab import EtherCrab, EtherCrabOption, Status
from pyautd3_link_ethercrab.native_methods.autd3_capi_link_ethercrab import NativeMethods as NativeEtherCrab
from pyautd3_link_ethercrab.native_methods.autd3_capi_link_ethercrab import Status as Status_


def test_status():
    lost = Status.Lost()
    state_change = Status.StateChanged()
    err = Status.Error()
    resumed = Status.Resumed()

    assert lost == Status.Lost()
    assert state_change == Status.StateChanged()
    assert err == Status.Error()
    assert resumed == Status.Resumed()
    assert lost != state_change
    assert lost != err
    assert lost != Status_.Lost
    assert lost != resumed
    assert state_change != err
    assert state_change != lost
    assert state_change != Status_.StateChanged
    assert state_change != resumed
    assert err != lost
    assert err != state_change
    assert err != Status_.Error
    assert err != resumed
    assert resumed != lost
    assert resumed != state_change
    assert resumed != err
    assert resumed != Status_.Resumed

    status = Status.__private_new__(Status_.Lost, "lost")
    assert status == Status.Lost()
    assert str(status) == "lost"

    with pytest.raises(NotImplementedError):
        _ = Status()


def test_ethercrab_is_default():
    assert NativeEtherCrab().link_ether_crab_is_default(EtherCrabOption()._inner())


def test_ethercrab():
    def err_handler(dev: int, status: Status) -> None:
        print(f"Device[{dev}], status: {status}")

    _ = EtherCrab(err_handler, EtherCrabOption())



import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3capi_driver import Duration, ResultLink

from pyautd3_link_ethercrab.native_methods.autd3_link_ethercrab import Status


class EtherCrabOption(ctypes.Structure):
    _fields_ = [
        ("ifname", ctypes.c_char_p),
        ("dc_configuration_sync0_period", Duration),
        ("state_check_period", Duration),
        ("sync_tolerance", Duration),
        ("sync_timeout", Duration),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EtherCrabOption) and self._fields_ == other._fields_  # pragma: no cover

    def __hash__(self) -> int:
        return super().__hash__()  # pragma: no cover


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3_capi_link_ethercrab{bin_ext}"))

        self.dll.AUTDLinkEtherCrab.argtypes = [ctypes.c_void_p, ctypes.c_void_p, EtherCrabOption]
        self.dll.AUTDLinkEtherCrab.restype = ResultLink

        self.dll.AUTDLinkEtherCrabIsDefault.argtypes = [EtherCrabOption]
        self.dll.AUTDLinkEtherCrabIsDefault.restype = ctypes.c_bool

        self.dll.AUTDLinkEtherCrabStatusGetMsg.argtypes = [ctypes.c_uint8, ctypes.c_char_p]
        self.dll.AUTDLinkEtherCrabStatusGetMsg.restype = ctypes.c_uint32

    def link_ether_crab(self, err_handler: ctypes.c_void_p, err_context: ctypes.c_void_p, option: EtherCrabOption) -> ResultLink:
        return self.dll.AUTDLinkEtherCrab(err_handler, err_context, option)

    def link_ether_crab_is_default(self, option: EtherCrabOption) -> ctypes.c_bool:
        return self.dll.AUTDLinkEtherCrabIsDefault(option)

    def link_ether_crab_status_get_msg(self, src: Status, dst: bytes) -> ctypes.c_uint32:
        return self.dll.AUTDLinkEtherCrabStatusGetMsg(src, dst)
